/*
Author : Benot Vogel -- BASED ON A 'BLANK' PLUGIN FROM THE ISADORA 1.3 SDK PROJECT WIZARD
Date : 24/05/2013
*/


// ===========================================================================
//	Isadora Demo Plugin			   2003 Mark F. Coniglio. All rights reserved.
// ===========================================================================
//
//	IMPORTANT: This source code ("the software") is supplied to you in
//	consideration of your agreement to the following terms. If you do not
//	agree to the terms, do not install, use, modify or redistribute the
//	software.
//
//	Mark Coniglio (dba TroikaTronix) grants you a personal, non exclusive
//	license to use, reproduce, modify this software with and to redistribute it,
//	with or without modifications, in source and/or binary form. Except as
//	expressly stated in this license, no other rights are granted, express
//	or implied, to you by TroikaTronix.
//
//	This software is provided on an "AS IS" basis. TroikaTronix makes no
//	warranties, express or implied, including without limitation the implied
//	warranties of non-infringement, merchantability, and fitness for a 
//	particular purpurse, regarding this software or its use and operation
//	alone or in combination with your products.
//
//	In no event shall TroikaTronix be liable for any special, indirect, incidental,
//	or consequential damages arising in any way out of the use, reproduction,
//	modification and/or distribution of this software.
//
// ===========================================================================
//
// CUSTOMIZING THIS SOURCE CODE
// To customize this file, search for the text ###. All of the places where
// you will need to customize the file are marked with this pattern of 
// characters.
//
// ABOUT IMAGE BUFFER MAPS:
//
// The ImageBufferMap structure, and its accompanying functions,
// exists as a convenience to those writing video processing plugins.
//
// Basically, an image buffer contains an arbitrary number of input and
// output buffers (in the form of ImageBuffers). The ImageBufferMap code
// will automatically create intermediary buffers if needed, so that the
// size and depth of the source image buffers sent to your callback are
// the same for all buffers.
// 
// Typically, the ImageBufferMap is created in your CreateActor function,
// and dispose in the DiposeActor function.

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

#include "IsadoraPluginPrefix.h"

#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"
#include "ImageBufferUtil.h"
#include "PluginDrawUtil.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>

// PLUGIN SPECIFIC INCLUDES
#include "ReadTextFromFile.h"

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN

	#include <windows.h>
	
	#define EXPORT_ __declspec(dllexport)
	
	#ifdef __cplusplus
	extern "C" {
	#endif

	BOOL WINAPI DllMain ( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpvReserved );

	#ifdef __cplusplus
	}
	#endif

	BOOL WINAPI DllMain (
		HINSTANCE	/* hInst */,
		DWORD		wDataSeg,
		LPVOID		/* lpvReserved */)
	{
	switch(wDataSeg) {
	
	case DLL_PROCESS_ATTACH:
		return 1;
		break;
	case DLL_PROCESS_DETACH:
		break;
		
	default:
		return 1;
		break;
	}
	return 0;
	}

#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------


// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

// Example: static int gMyGlobalVariable = 5;

// Possible strings for 'read mode' and 'new line mode' input properties

static const char* sReadModeStrings[5] = { "all", "char", "line", "paragraph", "undefined" };

static const char* sNewLineModeStrings[4] = { "windows", "unix", "mac", "undefined" };

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	MessageReceiverRef		mMessageReceiver;	// pointer to our message receiver reference
	
	
	String*					mPath;				// input text file path (path+name

	F_READMODE				mReadMode;			// F_ALL = all (default), F_CHR = character, F_LINE = line, F_PARAGRAPH = paragraph
	int						mStart;				// where the plugin should start reading (nth char/nth char/nth line/nth paragraph, respectively).

	bool					mAutoRead;			// true if auto-read is on. When false, the user has to user the read trigger manually when he wants to read the file and output its text.

	Boolean					mFileOK;			// true if chosen file exists and can be opened with 'read' attribute

	char*					mAllocatedString;	// string used to store the content of the entire text file
	char*					mFinalString;		// final string for output text. Only contains text filtered according to readmode and start parameters.

	bool					mBeginningFound;	// Flag set to true if beginning of final selection has been found
	bool					mEndFound;			// Flag set to true if end of final selection has been found

	int						mNewLineMode;		// F_WINDOWS for CR+LF text files, F_UNIX for LF, F_MAC for CR
	char					mNewLineChar;		// '\r' for mac, '\n' for windows and unix ('\r's in windows' \r\n are processed separately in the algorithm)
	
} PluginInfo;


// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupControl;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('PsIO');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "Read Text From File";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"INPROP file_path		path	string		text				-100	100		file.txt\r"

#if TARGET_OS_MAC
	"INPROP linebreak_mode	lbmd	int			propdef				0		2		2\r"
#elif TARGET_OS_WIN
	"INPROP linebreak_mode	lbmd	int			propdef				0		2		0\r"
#endif

	"INPROP read_mode		rdmd	int			propdef				0		3		0\r"
	"INPROP start			strt	int			number				0		*		0\r"
	"INPROP auto_read		autr	bool		onoff				0		1		0\r"
	"INPROP reset_output	rsto	bool		trig				0		1		0\r"
	"INPROP	read			trgr	bool		trig				0		1		0\r"

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
	"OUTPROP text_out		outt	string		text				*		*		\r"
	"OUTPROP file_available	fexs	bool		onoff				*		*		0\r"
	"OUTPROP has_been_read	fred	bool		trig				*		*		0\r";

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputPath = 1,
	kInputNewLineMode,
	kInputReadMode,
	kInputStart,
	kInputAutoRead,
	kInputResetOutput,
	kInputTriggerRead,
	
	kOutText = 1,
	kFileFound,
	kFileIsRead
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	// ACTOR HELP
	"This actor reads a text file and outputs its text (after optional filtering).\n\n"
		"IN:\n- File path\n- New lines system\n- Read mode\n- Auto-read file on parameter change on/off\n- Reset output\n- Trigger read\n\nOUT:\n- Output text\n- File can be read\n- 'File has been read' trigger"
		"\n\nSpecial characters will be kept, especially new lines (LF, CRLF...). This means that the 'new lines mode' input property must be chosen according to the system the text file has been edited on. "
		" Text files can be converted to other new lines systems using advanced text editors, such as Notepad++ (Windows), or shell scripts.",

	// INPUT HELP
	"The path to the file you want to read. Line breaks will be ignored, so feel free to cut long paths into several lines.\n\nExample: C:\\file.txt\nor:\nC:\\long\npath\\file.txt",
	
	"New line mode. Set this to 'windows' if the given text file uses CR+LF line breaks, 'unix' for LFs, 'mac' for CRs. This property is used for reading the file properly AND for ouput text formatting.",

	"Read mode.\n- all = read until end of file is reached.\n- char = read only 1 character.\n- line = read only 1 line of text (a blank line also counts as a line).\n- paragraph = read only 1 paragraph (paragraphs are separated by 2 blank lines or more).",

	"This parameter depends on the 'read mode' parameter. When 'read mode' is set to\n- all: read from the nth character\n- char: read only the nth character\n"
		"- line: read the nth line\n- paragraph: read the nth paragraph\n\nThis parameter starts from 0 : first position is 0, not 1.",

	"Auto-read mode.\n\nWhen set to on, the input file will be read and the output text will be updated whenever the 'file path', 'read mode', or 'start' input properties change.\nWhen set to off, you will have to use the 'read' input trigger each time you want to read the file.",

	"Clears the output string and resets the 'file available' property to off.",

	"Trigger reading. Read file and send its text to output. 'file available' and 'has been read' output properties will be set too if the file could be read.",

	// OUTPUT HELP

	"The output text. Special characters will be kept as they appear in the input file (be sure to select proper line break mode).",

	"'on' if input file can be opened and read. 'off' if not (this usually means that the input path is wrong).",

	"Triggered if input file has been read and output text has changed accordingly."
};

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation and initialization of private member variables
	
	// mAllocatedString (string corresponding to the entire text file content) first allocation
	info->mAllocatedString = (char*)malloc(sizeof(char));

	// mFinalString (string corresponding to filtered text only) first allocation
	info->mFinalString = (char*)malloc(sizeof(char));

	// use default line ending for current platform
	#if TARGET_OS_MAC
	info->mNewLineMode = F_MAC;
	#elif TARGET_OS_WIN
	info->mNewLineMode = F_WINDOWS;
	#endif

	// By default, read the entire file
	info->mReadMode = F_ALL;
	
	// Initialize the starting line with the very 1st line of the file
	info->mStart = 0;

	// Initialize auto-read mode to off
	info->mAutoRead = false;
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != nil);
	
	// ### destruction of private member variables
	
	// free mAllocatedString & mFinalString
	free(info->mAllocatedString);
	free(info->mFinalString);

	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}


// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//



// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = nil;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	


// ---------------------------------------------------------------------------------
//	###	 PropertyValueToString
// ---------------------------------------------------------------------------------
//	### This function defines custom display strings for "read mode" input parameter
//	### since we declared this parameter as "PROPDEF" type

static Boolean
PropertyValueToString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,				// property type
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			inValue,					// the property's current value
	char*				outString)					// string to display (max length of 256 characters)
{
	Boolean outStringChanged = false; // I don't know if it's necessary, let's assume it is (function declared with Boolean return)

	if (inPropertyType == kInputProperty) // input parameters only
	{

		switch (inPropertyIndex1) {

		case kInputReadMode : // 'read mode' input property
			
			switch (inValue->u.ivalue) {
				
			case F_ALL :
				strcpy(outString, sReadModeStrings[F_ALL]);
				outStringChanged = true;
				break;
			case F_CHAR :
				strcpy(outString, sReadModeStrings[F_CHAR]);
				outStringChanged = true;
				break;
			case F_LINE :
				strcpy(outString, sReadModeStrings[F_LINE]);
				outStringChanged = true;
				break;
			case F_PARAGRAPH :
				strcpy(outString, sReadModeStrings[F_PARAGRAPH]);
				outStringChanged = true;
				break;
			default :
				strcpy(outString, sReadModeStrings[F_UNDEFINED]);
				outStringChanged = true;
				break;

			}
			break;

		case kInputNewLineMode : // 'line break mode' input property
			
			switch (inValue->u.ivalue) {
				
			case F_WINDOWS :
				strcpy(outString, sNewLineModeStrings[F_WINDOWS]);
				outStringChanged = true;
				break;
			case F_UNIX :
				strcpy(outString, sNewLineModeStrings[F_UNIX]);
				outStringChanged = true;
				break;
			case F_MAC :
				strcpy(outString, sNewLineModeStrings[F_MAC]);
				outStringChanged = true;
				break;
			default :
				strcpy(outString, sNewLineModeStrings[F_NLUNDEFINED]);
				outStringChanged = true;
				break;

			}
			break;

		default :
			break;
		}
	}

	if (outStringChanged)
		return true;
	else
		return false;

}

// ---------------------------------------------------------------------------------
//		 PropertyStringToValue
// ---------------------------------------------------------------------------------

static Boolean
PropertyStringToValue(
	IsadoraParameters*	ip,
	ActorInfo*			/* inActorInfo */,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	const char*			inString,
	ValuePtr			outValue)
{
	Boolean encodeSuccessful = false;
	
	if (inPropertyType == kInputProperty && inPropertyIndex1 == kInputNewLineMode) {
		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '2') {
			outValue->type = kInteger;
			outValue->u.ivalue = (SInt32) inString[0] - '0';
			encodeSuccessful = true;
		} else {
			SInt32 matchIndex = LookupPartialStringInList_(ip, 3, sNewLineModeStrings, inString);
			if (matchIndex >= 0) {
				outValue->type = kInteger;
				outValue->u.ivalue = matchIndex;
				encodeSuccessful = true;
			}
		}
	}
	
	if (inPropertyType == kInputProperty && inPropertyIndex1 == kInputReadMode) {
		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '3') {
			outValue->type = kInteger;
			outValue->u.ivalue = (SInt32) inString[0] - '0';
			encodeSuccessful = true;
		} else {
			SInt32 matchIndex = LookupPartialStringInList_(ip, 4, sReadModeStrings, inString);
			if (matchIndex >= 0) {
				outValue->type = kInteger;
				outValue->u.ivalue = matchIndex;
				encodeSuccessful = true;
			}
		}
	}
	
	return encodeSuccessful;
}



// ---------------------------------------------------------------------------------
//		READTEXTFILE FUNCTION
// ---------------------------------------------------------------------------------
//	This function reads the text file according to the actor's variables. Those must be already set.
//

static void
ReadTextFile(
	IsadoraParameters*	ip,
	PluginInfo*			info)
{

	// trigger file reading and save it into the 'selected text' string
			
	FILE* fileToRead;			// file pointer
	char* filePath;				// file path string

	size_t fileSize = -1;		// file size (from stat() function)
	bool fileIsOpen = false;	// 'file is open' flag

	char* rmStart = nil;		// 1st character of the text we want to keep
	char* rmEnd = nil;			// last character of the text we want to keep (should point to '\0' at the end of the process)
			
	int curLine;				// line counter
	int curParagraph;			// paragraph counter
	bool newLineFound = false;	// this flag will be set to true if a new line has been detected while parsing info->mAllocatedString

	char* slidingStringPtr;		// temporary sliding pointer used to parse the string

	Value kOutTextValue = { kString, nil };				// Output text as a Value
	Value kFileIsAccessibleValue = { kBoolean, nil };	// 'File is accessible' boolean Value

	// First, let's find out if the file path is an empty string:
	// in this case we should do absolutely nothing in this function
	
	if (*(info->mPath->strData) == '\0')
		return;

	// FORMAT FILE PATH WITH DOUBLE BACKSLASHES, AND REMOVE POSSIBLE CARRIAGE RETURNS AND LINE FEEDS
	
	// count how many backslashes there are in the path string, in order
	// to be able to build a new string with doubled backslashes

	int numberOfBackslashes = 0; // counter

	const char *slidingPathReader = info->mPath->strData;

	#if TARGET_OS_MAC
	const bool runningOnWindows = false;
	#elif TARGET_OS_WIN
	const bool runningOnWindows = true;
	#endif
	
	// if we're running on windows, count the number
	// of backslahes because we will have to allocate
	// space to double them
	if (runningOnWindows) {
		do {
			if (*slidingPathReader == '\\')
				numberOfBackslashes++;
		} while (*(slidingPathReader++));
	}

	size_t formattedPathSize = strlen(info->mPath->strData) - 1 + numberOfBackslashes; // get total double-backslashed-path length

	// build the new path string

	filePath = (char*) malloc (formattedPathSize * sizeof(char));

	slidingPathReader = info->mPath->strData;

	char* slidingPathWriter = filePath;

	do {
		
		// if we're running on windows and we find a
		// backslash, we need to double it for fopen
		
		if (runningOnWindows && *slidingPathReader == '\\') {
			
			*slidingPathWriter = '\\';
			slidingPathWriter++;
			*slidingPathWriter = '\\';
			slidingPathWriter++;
		
		} else {
			
			// ignore '\r' and '\n' characters in the path name
			if (*slidingPathReader != '\r' && *slidingPathReader != '\n') {
			
				// at this point we know that the char is not '\\', '\r' nor '\n',
				// so it's a normal char (or '\0'): just copy it and go on
				*slidingPathWriter = *slidingPathReader;
				slidingPathWriter++;
			}
		}
	
	} while (*(slidingPathReader++)); // Read until '\0' has been read and copied

	// TRY TO OPEN THE FILE AND GET fopen STATUS
	#if TARGET_OS_MAC
	
		// on macOS, we can use the UTF8 path
		fileToRead = fopen(filePath, "r"); // try to open the file using the read attribute
	
	#elif TARGET_OS_WIN
	
		// on Windows, we must convert the path to a wide character string
	
		// determine number of WCHARs required and allocate buffer
		int count = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, filePath, (int) strlen(filePath), NULL, 0);
		size_t bufferSize = count + 1;
		WCHAR* wstr = (WCHAR*) malloc(bufferSize * sizeof(WCHAR));
		
		// attempt to convert UTF-8 path to WCHAR string
		int conversionResult = ::MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, filePath, (int) strlen(filePath), wstr, (int) bufferSize);
		if (conversionResult >= 0 && conversionResult < bufferSize) {
			wstr[conversionResult] = 0;
		} else {
			wstr[0] = 0;
		}

		fileToRead = _wfopen(wstr, L"r"); // try to open the file using the read attribute
	
		free(wstr);

	#endif

	if (fileToRead == NULL) { // test the result
		fileIsOpen = false;
	} else {
		fileIsOpen = true;
	}

	/*
	- SET VALUE FOR "file available" OUTPUT PARAMETER
	- IF FILE IS OPEN, READ IT AND STORE ITS CONTENTS TO A STRING
	- FIND BEGINNING AND END OF SELECTION ACCORDING TO READMODE INPUT PROPERTY
	- CLOSE THE FILE
	*/

	// test if file has been found and can be read

	if (fileIsOpen == false) {
		
		// prepare "file available" output param --> can't be opened in read mode
		info->mFileOK = false;

		// but do nothing else


	} else {
	
		// prepare "file available" output param --> has been opened in read mode
		info->mFileOK = true;

		// READ EVERY CHARACTER IN THE FILE AND STORE ALL OF THEM INTO info->mAllocatedString

		int currentChar;

		#if TARGET_OS_MAC
		struct stat fileStat;
		int stat_result = stat(filePath, &fileStat);
		#elif TARGET_OS_WIN
		struct _stat fileStat;
		int stat_result = _stat(filePath,&fileStat);
		#endif
		
		if (stat_result < 0) { // get file information
			
			// no Isadora alert function at this time
			printf("Error: file size cannot be determined using stat().\r\n");
			// error but we already know that the file is open, so this case should not happen
		
		} else {
					
			fileSize = fileStat.st_size; // get size

			info->mAllocatedString = (char*)realloc (info->mAllocatedString, (fileSize + 1 ) * sizeof(char)); // + 1 for null character at the end because the file is open in binary mode (not sure but seems to be OK)
				
			if (info->mAllocatedString == NULL) {
				printf("Memory allocation error while reallocating mAllocatedString.\r\n"); // No Isadora alert function at this time
				// no other error should occur here, as fileSize should never be -1
			}

			// write entire file to info->mAllocatedString

			char* storagePtr = info->mAllocatedString;

			while ( (currentChar = fgetc(fileToRead)) != EOF) {
				*storagePtr = currentChar;
				storagePtr++;
			}

			// Add string ending char
			*storagePtr = '\0';

			// FILTER STRING ACCORDING TO "READMODE" AND "START" PARAMETERS

			// find beginning and end according to what we want to keep in the string //
					
			// initialize beginning/end found flags
			info->mBeginningFound = false;
			info->mEndFound = false;

			// if there are blank lines at the beginning, ignore them first (read on
			// until 1st non-linebreak char is reached)

			slidingStringPtr = info->mAllocatedString;

			while ((*slidingStringPtr == '\r') || (*slidingStringPtr == '\n')) {
				slidingStringPtr++;
			}

			// if end of string has already been reached (=info->mAllocatedString
			// is empty or only contains new line characters) [this shouldn't happen
			// normally, but who knows if the file is empty or not?] an empty string
			// can already be set for selection
			
			if (*slidingStringPtr == '\0') {

				// Set beginning of selection
				rmStart = slidingStringPtr;
							
				// Set end of selection
				rmEnd = slidingStringPtr;

				// Set mBeginningFound & mEndFound flag in order to stop searching
				info->mBeginningFound = true;
				info->mEndFound = true;
							
			}

			switch (info->mReadMode) { // from now on the algorithm differs according to 'read mode' input property

			case F_UNDEFINED:
				break;
				
			case F_ALL :
						
				// set the bounds if they are not set yet //

				slidingStringPtr = info->mAllocatedString;
						
				if (info->mBeginningFound == false) {

					int i = 0;

					// read until start character is reached
					while (i < info->mStart) {
								
						// note: i = info->mStart && *slidingStringPtr == '\0' at the same
						// time can't happen here because this case has already been tested

						// if 'start' input property is larger than number of characters in the file
						if (*slidingStringPtr == '\0') {

							// set end of selection (beginning will be set after while loop) and associated flag
							rmEnd = slidingStringPtr;
							info->mEndFound = true;

							// force end of while loop (a bit nasty but it works)
							i = info->mStart;

						} else {
						
							// end of string has not been reached yet, just read on
							slidingStringPtr++;
							i++;

						}					
					}
							
					rmStart = slidingStringPtr;		// set beginning of selection
					info->mBeginningFound = true;	// and corresponding flag
				}

				// if '\0' char has not been reached yet, read on until we find it
				if (info->mEndFound == false) {

					while(*slidingStringPtr != '\0')
						slidingStringPtr++;

					// and set end of selection & corresponding flag
					rmEnd = slidingStringPtr;

					info->mEndFound = true;							
				}
						
				break;

			case F_CHAR :
						
				// extract only 1 char
						
				// set the bounds if they are not set yet

				slidingStringPtr = info->mAllocatedString;
						
				if (info->mBeginningFound == false) {

					int i = 0;

					while (i < info->mStart) { // read until start character is reached
								
						// note : i = info->mStart && *slidingStringPtr == '\0' at the same
						// time can't happen here because this case has already been tested

						if (*slidingStringPtr == '\0') { // if 'start' input property is larger than number of characters in the file

							// set end of selection (beginning will be set after while loop) and associated flag
									
							rmEnd = slidingStringPtr;
							info->mEndFound = true;

							// force end of while loop (a bit nasty but it works)
							i = info->mStart;

						} else {

							slidingStringPtr++;
							i++;

						}					
					}
							
					rmStart = slidingStringPtr; // set beginning of selection
					info->mBeginningFound = true; // and corresponding flag
				}

				// if '\0' char has not been reached yet, read only 1 char
				if (info->mEndFound == false) {
							
					// set end of selection & corresponding flag
					rmEnd = rmStart + sizeof(char);

					info->mEndFound = true;							
				}
						
				break;

			case F_LINE :

				// find beginning according to the line number ('start' input property)
				slidingStringPtr = info->mAllocatedString;

				curLine = 0;
				newLineFound = false;
						
				if (info->mBeginningFound == false) {

					// Initialize beginning of selection in case the 'start' property is 0
					rmStart = slidingStringPtr;

					// if 'start' property is greater than 0
					while (curLine < info->mStart) {


						if (*slidingStringPtr == '\0') {

							// set end of selection & corresponding flag
							rmEnd = slidingStringPtr;
							info->mEndFound = true;
							
							// force exit of while loop (a bit nasty but it works...)
							curLine = info->mStart;

						} else {

							// detect linebreaks
							if (*slidingStringPtr == info->mNewLineChar)
								newLineFound = true;				

							// read on
							slidingStringPtr++;

						}

						if (newLineFound) {

							// Increase current line no
							curLine++;

							// reset flag to be able to find new new-line characters from now on
							newLineFound = false;

						}
								
						// Set start of selection
						rmStart = slidingStringPtr;
					}

					info->mBeginningFound = true;
				}


				// if '\0' char has not been reached yet, read only 1 line
				if (info->mEndFound == false) {

					// find line end from here

					while ((*slidingStringPtr != '\r') && (*slidingStringPtr != '\n') && (*slidingStringPtr != '\0')) {
						slidingStringPtr++;
					}

					// Set end of selection & corresponding flag
					rmEnd = slidingStringPtr;

					info->mEndFound = true;

				}

				break;

			case F_PARAGRAPH :

				// find beginning according to the paragraph number ('start' input property)
				slidingStringPtr = info->mAllocatedString;

				char previousChar = NULL; // initialize previous character
				curParagraph = 0;
						
				if (info->mBeginningFound == false) {

					// initialize beginning of selection in case the 'start' property is 0
					rmStart = slidingStringPtr;

					// if 'start' property is greater than 0
					while (curParagraph < info->mStart) {

						if (*slidingStringPtr == '\0') {

							// Set end of selection & corresponding flag
							rmEnd = slidingStringPtr;

							info->mEndFound = true;

							curParagraph = info->mStart; // Force exit of while loop (a bit nasty but it works)


						} else {

							// detect single new lines
							if (*slidingStringPtr == info->mNewLineChar) {

								// If previous char was already a new line char
								if (previousChar == info->mNewLineChar)
								{
									// ... the end of a paragraph has been reached

									// Increase current paragraph #
									curParagraph++;

									// Read on until next non-new-line char is found (can be the beginning of a new paragraph or just '\0')
									while (*slidingStringPtr == '\r' || *slidingStringPtr == '\n') {
										slidingStringPtr++;
									}

									slidingStringPtr--; // just because we move on to the next char once again using *slidingStringPtr++ below, we should move one step backwards first
								}


							}

							// READ ON
							slidingStringPtr++;
						}

						// Set start of selection
						rmStart = slidingStringPtr;

						previousChar = *(slidingStringPtr-1); // Set previous char in order to be able to detect new paragraphs, not only single line endings

					}

					info->mBeginningFound = true;
				}

				if (info->mEndFound == false) { // if '\0' char has not been read yet, read only 1 paragraph (or until 'end of string' char is reached)

					// find paragraph end from here

					bool paragraphEndReached = false;

					while (*slidingStringPtr != '\0' && paragraphEndReached == false) {
							
						if (*slidingStringPtr == info->mNewLineChar) { // If current char is a new line char

							// test next chars to detect end of paragraph: if next char != '\0' and next char == new line char
							if (*(slidingStringPtr+1) != '\0') {

								if (*(slidingStringPtr+1) == info->mNewLineChar) {
									paragraphEndReached = true;
								}

							}
						}

						slidingStringPtr++;

					}

					// do not step back to last char if no paragraph should be read (this is to
					// prevent a bug when the file end is composed of several linebreaks)
					
					if (*rmStart != '\0') {

						// remove every '\r's and '\n's at the end of the selected paragraph

						while ( *(slidingStringPtr-1) == '\r' || *(slidingStringPtr-1) == '\n') {
							slidingStringPtr--;
						}

					}

					// Set end of selection & corresponding flag
					rmEnd = slidingStringPtr;

					info->mEndFound = true;

				}
				break;

			}

			// this should never happen, but just in case...
			if (info->mBeginningFound == false || info->mEndFound == false) {
				// ERROR : something went wrong
				printf("File reading error : Cannot build an output text from this file.\r\n"); // No Isadora alert function at this time
			}

		}
 
		// CLOSE THE FILE
		fclose(fileToRead);
				
		// BUILD NEW STRING info->mFinalString USING BEGINNING (rmStart) AND END (rmEnd) POINTERS FROM PREVIOUS FILTER
			
		// if windows CRLF mode, count how many linebreaks there are, as we need to allocate space for additional '\r's

		int windowsNewLinesCounter = 0;
				
		if (info->mNewLineMode == F_WINDOWS) {

			slidingStringPtr = rmStart;

			while (*slidingStringPtr != '\0') {
				if (*slidingStringPtr == info->mNewLineChar)
					windowsNewLinesCounter++;
				slidingStringPtr++;
			}
		}

		info->mFinalString = (char*)realloc(info->mFinalString, ((rmEnd - rmStart) + windowsNewLinesCounter +1 ) * sizeof(char)); // + 1 for '\0'

		if (info->mFinalString == NULL) {
			printf("Memory allocation error while reallocating mFinalString.\r\n"); // No Isadora alert function at this time
			// No other error should occur here, as size should never be 0
		}

		char* slidingFinalReader;
		char* slidingFinalWriter;

		slidingFinalReader = rmStart;
		slidingFinalWriter = info->mFinalString;

		while (slidingFinalReader < rmEnd) {

			// If CRLF mode, add '\r' before every '\n'
			if (info->mNewLineMode == F_WINDOWS) {
				if (*slidingFinalReader == info->mNewLineChar) {
					*slidingFinalWriter = '\r';
					slidingFinalWriter++;
				}
			}

			// copy current character
			*slidingFinalWriter = *slidingFinalReader;

			// increment read and write pointers
			slidingFinalReader++;
			slidingFinalWriter++;
		}

		*slidingFinalWriter = '\0'; // add string ending char (as it is not always the case)


		// SET ALL OUTPUT PARAMETERS FOR FEEDBACK (except 'file available' as it should
		// always be updated, not only when the file could be read)

		// OUTPUT TEXT FROM BUILT SELECTION

		AllocateValueString_(ip, info->mFinalString, &kOutTextValue);

		SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutText, &kOutTextValue);

		ReleaseValueString_(ip, &kOutTextValue);

		// "FILE READ" OUTPUT TRIGGER
			
		if (info->mFileOK == (Boolean)true) { // only if file could be opened
			Value kFileHasBeenReadValue = { kBoolean, nil };
			kFileHasBeenReadValue.u.ivalue = 1;
			SetOutputPropertyValue_(ip, info->mActorInfoPtr, kFileIsRead, &kFileHasBeenReadValue);
		}



	} // end if fopen is successful
	
	// SET FILE AVAILABLE ON/OFF OUTPUT PROPERTY

	kFileIsAccessibleValue.u.ivalue = info->mFileOK;

	SetOutputPropertyValue_(ip, info->mActorInfoPtr, kFileFound, &kFileIsAccessibleValue);

}




// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				/* inInitializing */)		// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	// ### When you add/change/remove properties, you will need to add cases
	// to this switch statement, to process the messages for your
	// input properties
	
	// The value comes to you encapsulated in a Value structure. See 
	// ValueCommon.h for details about the contents of this structure.
	
	Value kOutTextValue = { kString, nil }; // Output text as a Value
	Value kFileIsAccessibleValue = { kBoolean, nil }; // 'File is accessible' boolean Value
	
	Value pathValue = { kString, nil };
	char* slidingPathPtr;

	switch (inPropertyIndex1) {

	case kInputPath: // Input path changed
		
		/*
		Paths beginning with white spaces and/or linebreaks will bugs later on while trying to read the file.
		Therefore we should clean this path by trimming all white spaces and linebreaks at the beginning of the input path.

		(totally empty input path string will be tested and ignored).

		This operation should be done here so that the user gets feedback directly into his input property box.
		*/

		// Find the real beginning of the string (skip white spaces and linebreaks)
		slidingPathPtr = inNewValue->u.str->strData;
		
		while (*slidingPathPtr != '\0' && (*slidingPathPtr == '\r' || *slidingPathPtr == '\n' || *slidingPathPtr == ' ')) {
			slidingPathPtr++;
		}

		// If new path is different from typed input, change the path input property
		if (slidingPathPtr != inNewValue->u.str->strData) {
			
			// allocate a Value in order to set the new path
			AllocateValueString_(ip, slidingPathPtr, &pathValue);
			
			// recursion : calling SetInputPropertyValue_() will launch the cause HandlePropertyChangeValue
			// to be called recusrisvely, but with formatted path as an input this time
			SetInputPropertyValue_(ip, info->mActorInfoPtr, kInputPath, &pathValue);
			
			// when we exit it and come back here, we should dispose the string in the Value, as we don't need it anymore
			ReleaseValueString_(ip, &pathValue);

		// If the path input is already formatted properly (= has been typed in a valid manner OR
		// has been formatted during previous execution (before recursive call))
		} else {

			// Set new path using the input as it is
			info->mPath = inNewValue->u.str;
		}

		// if auto-read is on, launch read function
		if (info->mAutoRead)
			ReadTextFile(ip, info);

		break;


	// linebreak mode changed
	case kInputNewLineMode:

		// set new 'new line' mode (should not be necessary, but just in 'case' ^^)
		// and corresponding linebreak char (this is necessary)
		switch (inNewValue->u.ivalue) {
			
		case F_WINDOWS:
			info->mNewLineMode = F_WINDOWS;
			info->mNewLineChar = '\n';
			break;
		case F_UNIX:
			info->mNewLineMode = F_UNIX;
			info->mNewLineChar = '\n';
			break;
		case F_MAC:
			info->mNewLineMode = F_MAC;
			info->mNewLineChar = '\r';
			break;
		}
		break;


	case kInputReadMode: // Read mode changed

		// set new read mode (should not be necessary, but just in "case" ^^)
		switch (inNewValue->u.ivalue) {
		case F_ALL: info->mReadMode = F_ALL; break;
		case F_CHAR: info->mReadMode = F_CHAR; break;
		case F_LINE: info->mReadMode = F_LINE; break;
		case F_PARAGRAPH: info->mReadMode = F_PARAGRAPH; break;
		}

		// If auto-read is on, launch read function
		if (info->mAutoRead)
			ReadTextFile(ip, info);

		break;

	// starting char/line/paragraph changed
	case kInputStart:

		// set new section start for reading
		info->mStart = (int) inNewValue->u.ivalue;

		// If auto-read is on, launch read function
		if (info->mAutoRead)
			ReadTextFile(ip, info);

		break;

	// auto-read status changed
	case kInputAutoRead:

		// set new section start for reading
		if (inNewValue->u.ivalue == 0)
			info->mAutoRead = false;
		else
			info->mAutoRead = true;

		// Launch read function it auto-read is now on
		if (info->mAutoRead)
			ReadTextFile(ip, info);

		break;


	// reset output is triggered
	case kInputResetOutput:

		// update 'file available' output property (set to off)
		kFileIsAccessibleValue.u.ivalue = false;
		SetOutputPropertyValue_(ip, info->mActorInfoPtr, kFileFound, &kFileIsAccessibleValue);

		// clean output text property
		
		AllocateValueString_(ip, "", &kOutTextValue);
		SetOutputPropertyValue_(ip, info->mActorInfoPtr, kOutText, &kOutTextValue);
		ReleaseValueString_(ip, &kOutTextValue);

		break;

	// reading is initiated
	case kInputTriggerRead:

		// launch read function
		ReadTextFile(ip, info);

		break;
	}
}


// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	if (!gPictInfo.mInitialized) {
		// PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	// place picture in top area
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	
	// don't draw anything in bottom area
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kCurrentIsadoraCallbackVersion;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= NULL; // I removed this one, though it is marked as required, hope it's OK
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= PropertyValueToString;	// For read mode and linebreak mode input properties
	outActorParams->mPropertyStringToValueProc			= PropertyStringToValue;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

// ---------------------------------------------------------------------------------
//		 ProcessVideoFrame
// ---------------------------------------------------------------------------------
//	### This is the code that does the actual processing of a video frame. Modify
// this code to create your own filter.
//


// ---------------------------------------------------------------------------------
//		 ReceiveMessage
// ---------------------------------------------------------------------------------
//	Isadora broadcasts messages to its Message Receives depending on what message
//	they are listening to. In this case, we are listening for kWantVideoFrameTick,
//	which is broadcast periodically (30 times per second.) When we receive the
//	message, we check to see if our video frame needs to be updated. If so, we
//	process the incoming video and pass the newly generated frame to the output.

